/************************************************************************/
/*									*/
/*	server.c							*/
/*									*/
/*	Copyright (c) 1993, Research Systems Inc.  All rights reserved.	*/
/*									*/
/*	Run IDL as a child server process (IPC mechanism v. RPC).	*/
/*									*/
/*	Special commands:						*/
/*									*/
/*	go:	make IDL become interactive				*/
/*	quit:	exit							*/
/*									*/
/************************************************************************/

#include        <stdio.h>
#include        <ctype.h>
#include	<signal.h>
#include        <rpc/rpc.h>
#include	"export.h"
#include	"rpc_idl.h"
#include	<sys/time.h>

#ifdef ultrix
#define		XTERM		"/usr/bin/dxterm"
#else
#define		XTERM		"xterm"
#endif

/*	On some machines windowed subprocesses have developed a		*/
/*	resistance to SIGTERM.  Use stronger measures to insure		*/
/*	death.								*/

#if defined(sgi) || defined(ALPHA_OSF)
#define		DONE_SIGNAL	SIGKILL
#else
#define		DONE_SIGNAL	SIGTERM
#endif



static	CLIENT		* idl_id;	/* RPC server ID		*/
static	int		child_pid;	/* process ID of xterm/idl	*/
static	LONG		shm_done;	/* current command		*/

static	void		cleanup(int);
static	void		cmd_sig(int);

/*
 *  cmd_sig
 *	Server will send us a SIGUSR1 signal to indicate completion
 *	of a command.
 *	Note: using signal this way is the most portable but not the
 *	most robust or the best machine dependent solution.
 */
static void	cmd_sig( int sig )
{
    signal(sig, cmd_sig);		/* Continue to catch SIGUSR1	*/
    shm_done	= 0;			/* Clear command-done flag	*/
}


/*
 *  cleanup
 *	Signal processing routine.
 *	Free system resources.  Try to make sure IDL server
 *	cleans up too.
 */
static void	cleanup( int sig )
{
    kill(child_pid, DONE_SIGNAL);
    exit(0);
}


/*
 *  run_cmd
 *	Send a command to the server via RPC mechanism.
 *	Handle the different scenarios:
 *		Command completes.
 *		Command runs but RPC times out
 *			Wait for shared memory cmd-done-flag to clear
 *			Wait for server to signal us cmd-is-done
 *		RPC mechanism failed.
 */
void	run_cmd( char * cmd )
{
    int         	error_code;	/*	Server error code	*/
    struct rpc_err	err;		/*	RPC error code		*/

    /*  Sent the command to the server  */

    shm_done	= 1;
    if((error_code =  send_idl_command(idl_id, cmd)) < 0) {
	if( error_code == -1) {
	    /*	We could have had an error occur for many	*/
	    /*	reasons: If its a time out then wait for	*/
	    /*	the IDL server to contact us to determine	*/
	    /*	that the command is done. Otherwise, we have	*/
	    /*	a serious problem.				*/

	    CLNT_GETERR(idl_id, &err);
	    if(err.re_status == RPC_TIMEDOUT) {
		while(shm_done) {
		    sleep(1);
		}
	    }
	    else {
		printf("Server communication protocol failure\n");
		cleanup(0);
	    }
	}
    }
}

/*
 *  start_idl()
 *      creates an IDL server with a server id based on our process ID
 *      (attempt a unique number)
 */
static void	start_idl(void)
{
    int			i;
    int			retry;
    int			pid;
    LONG                server_id;
    char                cmd[MAX_STRING_LEN];

    /*	Compute a server id based on my process id.			*/
    pid		= getpid();
    server_id   = 0x2010CAFE + pid;

    /*	Make sure we clean up whenever the user quits, interrupts, etc	*/
    signal(SIGHUP, cleanup);
    signal(SIGINT, cleanup);
    signal(SIGQUIT, cleanup);
    signal(SIGTERM, cleanup);

    /*	Process command done signal					*/
    signal(SIGUSR1, cmd_sig);

    /*	Create the IDL process command.					*/
    sprintf(cmd, "exec %s -geometry 80x24 -e idl -server=%X %X",
	    XTERM, server_id, pid);

    switch( child_pid = fork() ) {
      case 0:					/*	Child process	*/
	execl("/bin/sh", "sh", "-c", cmd, 0);
      case -1:					/*	Fork error	*/
	fprintf(stderr,"Cannot fork an IDL server\n");
	exit(1);
      default:					/*	Parent process	*/
	set_rpc_verbosity(0);
	sleep(2);
	for(i=0;i<20;i++) {
	    if((idl_id = register_idl_client(server_id, NULL, NULL)) != NULL)
		return;
	    sleep(1);
	}
	fprintf(stderr, "Cannot register with IDL server");
	exit(1);
    }
}



/*
 *  main
 *	Start IDL server.
 *	Read a command, run a command
 */
public void main(int argc, char *argv[])
{
    char	* prompt = "RMTIDL> ";
    char	cmd[ MAX_STRING_LEN ];

    /*	Start IDL server					*/
    start_idl();

    printf("Type 'go' to make IDL process interactive\n");
    printf("Type ^C, ^D or 'quit' to quit this process\n\n");

    for(;;) {
	/*	Run a command loop				*/
	printf(prompt);
	fflush(stdout);
	
	cmd[0]	= (char)(-1);
	gets( cmd );		/* read a line from the user	*/

	if( cmd == '\0' )	/* ignore blank lines		*/
	    continue;
	
	/*	Check if user wants to quit:			*/
	/*	EOF or 'quit'					*/
	
	if( *cmd == (char)(-1) || !strcmp(cmd, "quit"))
	    break;


	/*	Check for going interactive			*/

	if( !strcmp(cmd, "go") ) {
	    shm_done	= 1;
	    idl_server_interactive( idl_id );
	    while(shm_done)
		sleep(1);
	}
	else
	    run_cmd( cmd );
    }
    printf("Exiting...\n");
    cleanup(0);
}
